#ifndef LAYER1_FREQ_EST_H
#define LAYER1_FREQ_EST_H

#include <complex.h>
#include <stdlib.h>

/*!
 * \brief Estimate the frequency of unknown BPSK symbols.
 *
 * This function squares all given symbols to remove data ambiguity, calculates
 * the phase values and then performs linear regression on those to determine
 * the frequency.
 *
 * \param recv        Pointer to the array of received symbols.
 * \param n           Number of symbols given.
 * \param final_phase Pointer to a float where the phase correction value is
 *                    stored. May be NULL if that value is not required.
 * \returns           The estimated frequency in radians per sample.
 */
float freq_est_unknown_bpsk(const float complex *recv, size_t n, float *final_phase);

/*!
 * \brief Estimate the frequency of known symbols.
 *
 * This function removes the data influence by calculating
 * recv[k]*conj(ref[k]), and then calls \ref freq_est_data_free() with the
 * result.
 *
 * \param recv        Pointer to the array of received symbols.
 * \param ref         Pointer to the array of reference symbols (e.g. a preamble).
 * \param n           Number of symbols given.
 * \param final_phase Pointer to a float where the phase correction value is
 *                    stored. May be NULL if that value is not required.
 * \returns           The estimated frequency in radians per sample.
 */
float freq_est_known_symbols(const float complex *recv, const float complex *ref, size_t n, float *final_phase);

/*!
 * \brief Estimate the frequency from data-free symbols.
 *
 * This function calculates the phase values from the given symbols and then
 * performs linear regression on those to determine the frequency.
 *
 * \param symbols     Pointer to the array of received symbols.
 * \param n           Number of symbols given.
 * \param final_phase Pointer to a float where the phase correction value is
 *                    stored. May be NULL if that value is not required.
 * \returns           The estimated frequency in radians per sample.
 */
float freq_est_data_free(const float complex *symbols, size_t n, float *final_phase);

#endif // LAYER1_FREQ_EST_H
