#ifndef LAYER1_TRANSMISSION_H
#define LAYER1_TRANSMISSION_H

#include <stdlib.h>
#include <complex.h>

#include <liquid/liquid.h>

#include "results.h"

typedef struct
{
	firinterp_crcf   rrc_interp;
} transmission_ctx_t;


/*!\brief Initialize the transmission object.
 *
 * \param ctx   Pointer to the context to initialize.
 * \returns     The result code of the initialization.
 */
result_t transmission_init(transmission_ctx_t *ctx);


/*!\brief Free all memory associated with the given context.
 *
 * \param ctx   Pointer to the context to free.
 * \returns     The result code of the operation.
 */
result_t transmission_free(transmission_ctx_t *ctx);


/*!\brief Generate a signal ramp-up and write it to the output array.
 *
 * \param ctx           Pointer to the transmission context.
 * \param output        Pointer to the output array.
 * \param output_size   Pointer to the output size. Must be set to the number
 *                      of samples that can be stored in \ref output when
 *                      calling the function. It will be changed to the number
 *                      of samples actually produced.
 * \returns             The result code of the operation.
 */
result_t transmission_ramp_up(transmission_ctx_t *ctx, float complex *output, size_t *output_size);


/*!\brief Generate a signal ramp-down and write it to the output array.
 *
 * \param ctx           Pointer to the transmission context.
 * \param output        Pointer to the output array.
 * \param output_size   Pointer to the output size. Must be set to the number
 *                      of samples that can be stored in \ref output when
 *                      calling the function. It will be changed to the number
 *                      of samples actually produced.
 * \returns             The result code of the operation.
 */
result_t transmission_ramp_down(transmission_ctx_t *ctx, float complex *output, size_t *output_size);


/*!\brief Filter the given symbols.
 *
 * \param ctx           Pointer to the transmission context.
 * \param packet        Pointer to the symbols to filter.
 * \param packet_size   Number of samples in \ref packet.
 * \param output        Pointer to the output array.
 * \param output_size   Pointer to the output size. Must be set to the number
 *                      of samples that can be stored in \ref output when
 *                      calling the function. It will be changed to the number
 *                      of samples actually produced.
 * \returns             The result code of the operation.
 */
result_t transmission_filter_packet(
		transmission_ctx_t *ctx,
		const float complex *packet,
		size_t packet_size,
		float complex *output,
		size_t *output_size);


/*!\brief Calculate the number of samples required for ramp-up.
 *
 * \param ctx           Pointer to the transmission context.
 * \returns             The number of samples required.
 */
size_t transmission_calculate_ramp_up_length(transmission_ctx_t *ctx);


/*!\brief Calculate the number of samples required for ramp-down.
 *
 * \param ctx           Pointer to the transmission context.
 * \returns             The number of samples required.
 */
size_t transmission_calculate_ramp_down_length(transmission_ctx_t *ctx);


/*!\brief Calculate the number of samples required for the given number of packet symbols.
 *
 * \param ctx           Pointer to the transmission context.
 * \param symbol_count  Size of the packet in symbols.
 * \returns             The number of samples required.
 */
size_t transmission_calculate_packet_length(transmission_ctx_t *ctx, size_t symbol_count);


#endif // LAYER1_TRANSMISSION_H
